(ns puppetlabs.rbac.schema
  (:require [schema.core :as sc]))

(def Action {:name          sc/Str
             :display_name  sc/Str
             :description   sc/Str
             :has_instances sc/Bool})

(def ObjectType {:object_type  sc/Str
                 :display_name sc/Str
                 :description  sc/Str
                 :actions      [Action]})

(def Permission {:object_type          sc/Str
                 :action               sc/Str
                 :instance             sc/Str})

(def PermissionWithID
  (assoc Permission
         (sc/optional-key :id) sc/Int))

(def BaseUser {:id           sc/Int
               :login        sc/Str
               :display_name sc/Str
               :email        sc/Str
               :last_login   (sc/maybe sc/Str)
               :role_ids     [sc/Int]
               :is_group     (sc/eq false)
               :is_revoked   sc/Bool})

(def LocalUser
  (assoc BaseUser
         :is_remote    (sc/eq false)
         :is_superuser sc/Bool))

(def RemoteUser
  (assoc BaseUser
         :group_ids          [sc/Int]
         :inherited_role_ids [sc/Int]
         :is_remote          (sc/eq true)
         :is_superuser       (sc/eq false)))

(def User
  (sc/if :is_remote RemoteUser LocalUser))

(def RemoteGroup {:id                 sc/Int
                  :login              sc/Str
                  :display_name       sc/Str
                  :role_ids           [sc/Int]
                  :user_ids           [sc/Int]
                  :is_group           (sc/eq true)
                  :is_remote          (sc/eq true)
                  :is_superuser       (sc/eq false)})

(def PostUser {:login        sc/Str
               :email        sc/Str
               :display_name sc/Str
               :role_ids     [sc/Int]})

(def PostGroup {:login    sc/Str
                :role_ids [sc/Int]})

(def ChangePassword {:current_password sc/Str
                     :password sc/Str})

(def ChangePasswordWithToken {:token sc/Str
                              :password sc/Str})

(def Role {(sc/optional-key :id) sc/Int
           :display_name         sc/Str
           :description          sc/Str
           :user_ids             [sc/Int]
           :group_ids            [sc/Int]
           :permissions          [PermissionWithID]})

;; It is valid to have an "empty" directory settings map; this means
;; the DS has been disconnected.
;; consider using "user-dn-template," like user-dn-template = uid={0},ou=users,o=myDn
(def maybe-str (sc/maybe sc/Str))
(def maybe-int (sc/maybe sc/Int))
(def maybe-bool (sc/maybe sc/Bool))

(def PermissionQuery
  {:object_type sc/Str
   (sc/optional-key :action) sc/Str
   (sc/optional-key :instance) sc/Str})

(sc/defn permission->string :- sc/Str
  "Convert a Permission map to a shiro permission string"
  [{:keys [object_type action instance]} :- Permission]
  (str object_type ":" action ":" instance))

(def PermissionsQuery
  {:token       sc/Str
   :permissions [PermissionQuery]})

(def PermittedQuery
  {:token sc/Str
   :permissions [Permission]})

(def DirectoryServiceSettings
  {(sc/optional-key :id) sc/Int

   ; KEY NAME              ; TYPE    ; EXAMPLE

   ; Meta
   :display_name           sc/Str ; AD
   :help_link              maybe-str ; http://initech.com/login-help
   (sc/optional-key :type) (sc/maybe (sc/enum "apacheds" "openldap" "activedirectory"))

   ; Connecting
   :hostname               sc/Str ; localhost
   :port                   sc/Int ; 10389
   :login                  sc/Str ; cn=admin,ou=users,dc=rg
   :password               sc/Str ; s3cr3t
   :ssl                    sc/Bool
   :connect_timeout        maybe-int ; 30

   ; Querying
   :base_dn                sc/Str ; dc=puppetlabs,dc=com

   :user_rdn               sc/Str ; ou=user
   :user_display_name_attr sc/Str ; displayName
   :user_email_attr        sc/Str ; mail
   :user_lookup_attr       sc/Str ; uid; must be uniquely identifying

   :group_rdn              sc/Str ; ou=groups
   :group_lookup_attr      sc/Str ; cn; must be uniquely identifying
   :group_object_class     maybe-str ; groupOfUniqueNames
   :group_name_attr        sc/Str ; cn
   :group_member_attr      sc/Str ; uniqueMember

   })
